--  2006 Peter Thiemann
{- | String-based, direct-style implementation of transactional database access.
-}
module Database.TRANSDB.Direct 
  (
    -- * Database connection
    D (), register,
    -- * Database tuple handles
    T (), findByPK, findByNV, find, getPrimaryKey,
    -- * Operations on tuples
    create, remove, 
    get, getMany, getAll, set, setMany, 
    -- * Transaction support
    with
  )
where

-- import qualified Database.TRANSDB.TraceQuery as DB

import qualified List
import Maybe (fromJust)
import IO
import Data.IORef


import Database.TRANSDB.Backend
import Database.TRANSDB.Manager
import Database.TRANSDB.LowLevel

import qualified Database.TRANSDB.SQL as SQL


-- the transactional database monad is a state monad lifted over @IO@ where the
-- state is the transaction log. Not using @StateT@ to avoid dependencies on
-- functional parameters etc.
data TDB b a =
    TDB { unTDB :: TState b -> IO (a, TState b)}

instance Monad (TDB b) where
  return x = TDB (\ s -> return (x, s))
  m >>= f  = TDB (\ s -> do (x', s') <- unTDB m s
			    unTDB (f x') s')

ioTDB :: IO a -> TDB b a
ioTDB action
  = TDB (\ s -> do x <- action
		   return (x, s))
  
data TState b =
    TState { getLog :: [LogEntry]
	   , getResult :: IORef b }

runTDB :: TState b -> TDB b a -> IO (a, TState b)
runTDB ts tdb = 
  unTDB tdb ts

applyToLogTDB :: ([LogEntry] -> [LogEntry]) -> TDB b [LogEntry]
applyToLogTDB g =
  TDB (\ ts -> let x = getLog ts 
		   y = g x
	       in  return (y, ts { getLog = y }))

getTheLogTDB :: TDB b [LogEntry]
getTheLogTDB = 
  applyToLog id

getTheRef :: TDB b (IORef b)
getTheRef =
  TDB (\ s -> return (getResult s, s))

class Monad t => TMonad t where
  io         :: IO a -> t a
  applyToLog :: ([LogEntry] -> [LogEntry]) -> t [LogEntry]
  getTheLog  :: t [LogEntry]

instance TMonad (TDB b) where
  io = ioTDB
  applyToLog = applyToLogTDB
  getTheLog = getTheLogTDB


-- |Create a fresh tuple in @table@ with new primary key and set its initial
-- contents. Returns a handle for the tuple. The transaction will not commit if
-- the initial contents do not fit (but it should better raise an exception
-- immediately). 
-- 
-- Complication: the fresh tuple must not be visible outside this transaction
-- but we need to ensure that its unique primary key is reserved, anyway. Hence,
-- we just reserve it through the sequence first time round.
create :: D -> TableName -> [NamedValue] -> TDB b T
create db table namedValues =
  do tid <- io $ do 
        conn <- getCachedConnection db
	pk <- generatePrimaryKey conn table
	return $ T db table pk
     applyToLog (NEW_TUPLE tid (List.sort namedValues) :)
     return tid


-- |Remove tuple identified by a primary key from DB. Only registers the action
-- in the log, but checks that the tuple exists.
-- Fails (with an exception) if the tuple does not exist.
remove :: T -> TDB b ()
remove tid =
  do tupleExists <- checkTupleExists tid False
     if tupleExists 
	then 
	  do applyToLog (REM_TUPLE tid :)
	     return ()
	else
	  fail ("DB.Transaction.remove: tuple " ++ show tid ++ " not present")

-- |Check existence of a tuple.
checkTupleExists :: T -> Bool -> TDB b Bool
checkTupleExists tid doRead =
  do log <- getTheLog
     let tupleIDStatus = checkTupleIDStatus tid log
     case tupleIDStatus of
       Exists ->
	 return True
       Removed ->
	 return False
       Unknown ->
	 if doRead 
	 then
	   do nvs <- io $ reallyReadValues tid
	      let pkColumn = primaryKeyColumn (t_tableName tid)
	      case lookup pkColumn nvs of
		Just (Just _) ->
		  do applyToLog (READ_TUPLE tid nvs [] [] :)
		     return True
		_ ->
		  return False
	 else
	   do exists <- io $ reallyCheckExists tid
	      return exists

data TupleIDExistence =
    Exists | Removed | Unknown

-- |does this tuple ID exist in the log?
checkTupleIDStatus :: T -> [LogEntry] -> TupleIDExistence
checkTupleIDStatus tid les =
  foldr g Unknown les
  where
    g :: LogEntry -> TupleIDExistence -> TupleIDExistence
    g le b =
      if le_tid le /= tid 
      then b
      else -- this log entry is for our tuple
      case le of
	REM_TUPLE {} ->
	  Removed
	_ ->
	  Exists

-- |Obtain a tuple through its primary key. Checks for its existence and
-- preloads the tuple in the cache.
findByPK :: D -> TableName -> PrimaryKey -> TDB b (Maybe T)
findByPK db table pk =
  let tid = T db table pk in
  checkTupleExists tid True >>= \ tupleExists ->
  if tupleExists 
  then return (Just tid)
  else return Nothing
  
-- |Obtain tuples through a where-clause of the form
-- name_i=value_i. Preloads. Caution: may generate an unlimited number of log
-- entries. 
-- 
-- TODO: Should perhaps take a @LIMIT@ parameter
findByNV :: D -> TableName -> [NamedValue] -> String -> TDB b [T]
findByNV db table nvs extraSQL =
  do allReads <- io $ reallyReadTuples db table [] nvs extraSQL
     applyToLog (allReads ++)
     return (map le_tid allReads)

-- |Take a database, a table, and an SQL where-clause. 
-- Return a list of tuple IDs for the results.
find :: D -> TableName -> WhereClause -> TDB b [T]
find db table whereClause =
  do pks <- io (findIO db table whereClause)
     let makeT pk = T db table pk
     return $ map makeT pks

-- |Get the primary key from a tuple handle. Does not access the DB at all. Does
-- not affect the transactional behavior of the tuple (it does not count as a
-- read operation on the tuple).
getPrimaryKey :: T -> DBValue
getPrimaryKey tid =
  Just (t_primaryKey tid)

-- |Get field value from handle. Caches the whole tuple.
get :: T -> ColumnName -> TDB b (DBValue)
get tid field =
  do [value] <- getMany tid [field]
     return value

-- |Get field values from handle. Caches the whole tuple.
getMany :: T -> [ColumnName] -> TDB b [DBValue]
getMany tid colNames =
  do nvs <- getInternal tid (readValues' tid colNames)
     return (map snd nvs)

getInternal :: 
  T -> ([LogEntry] -> IO ([NamedValue], [LogEntry])) -> TDB b [NamedValue]
getInternal tid readValues =
  do log <- getTheLog
     (values, newEntries) <- io $ readValues log
     applyToLog (newEntries ++)
     return values

-- |Get whole tuple from handle.
getAll :: T -> TDB b [NamedValue]
getAll tid =
  getInternal tid (readAllValues tid)

-- |Set field value through handle.
set :: T -> ColumnName -> DBValue -> TDB b ()
set handle field val =
  -- trivial implementation
  setMany handle [(field, val)]

-- |Set many columns at once through handle
setMany :: T -> [NamedValue] -> TDB b ()
setMany tid namedValues =
  do applyToLog (SET_TUPLE tid namedValues :)
     return ()

-- |Abstract of the nested vs the unnested case. @target@ is the target
-- monad. It may be @IO@ (the unnested case) or @TDB b@ (the nested case).
class Monad target => With target where
  with :: 
     result				-- default result
     -> TDB result result
     -> target result

instance With IO where
  with = withIO

instance With (TDB b) where
  with = withTDB

-- |abandons the innermost active transaction and makes it return its @result@
-- argument. 
abandon :: result -> TDB result ()
abandon res =
  do setFail res
     io $ ioError (userError ":abandon:")

-- |sets the return value of the innermost active transaction for the case that
-- this transaction fails to commit.
setFail :: result -> TDB result ()
setFail res =
  do r <- getTheRef
     io (writeIORef r res)

-- |Opens a transactional scope for the unnested case. Takes a default @result@
-- value and a transaction with the same @result@ type and constructs an @IO@
-- action returning the same value of the transaction (if that commits
-- successfully). The default value is returned if the transaction fails without
-- explicitly setting a result value (using 'setFail').
withIO :: result -> TDB result result -> IO result
withIO result body =
  do resultOnFail <- newIORef result
     let
	initialstate = 
	  TState { getLog = []
		 , getResult = resultOnFail 
		 }
     lr <- try (runTDB initialstate body)
     case lr of
      Right (v, finalstate) ->
	do let log = getLog finalstate
	   committed <- commitFromLog log
	   if committed then
	      return v
	    else do
	      res <- readIORef resultOnFail
	      return res
		 
      Left err ->
		 -- Caught an exception; might be due to an @anbandon@ operation
	if ioeGetErrorString err == ":abandon:" then
	   readIORef resultOnFail
	 else
	   -- might want to commit the transaction first?
	   ioError err

-- |Opens a nested transactional scope. Same functionality as 'withIO'. However,
-- a nested transaction cannot fail to commit because it does not check its
-- preconditions against the outside world. Rather the nested transaction only
-- manipulates the log.
withTDB :: a -> TDB a a -> TDB b a
withTDB result body =
  do initialLog <- getTheLog
     resultOnFail <- io (newIORef result)
     let 
       initialstate =
	 TState { getLog = initialLog
		, getResult = resultOnFail
		}
     lr <- io $ try (runTDB initialstate body)
     case lr of
      Right (v, finalstate) ->
	do let log = getLog finalstate
	   applyToLog (const log)
	   return v
	   -- cannot fail to commit at this point!
		 
      Left err ->
	io $	 -- Caught an exception; might be due to an @anbandon@ operation
	if ioeGetErrorString err == ":abandon:" then
	   -- just continue with the old log in outer transaction
	   readIORef resultOnFail
	 else
	   -- might want to commit the transaction first?
	   ioError err
