--  2006 Peter Thiemann
{- | low-level functions for transaction implementation.
-}
module Database.TRANSDB.LowLevel
  (
    NamedValue, WhereClause, DBValue,
    LogEntry (..), T (..),
    reallyReadValues, reallyCheckExists, reallyReadTuples, findIO, 
    readValues', readAllValues, checkFieldsForExistence,
    commitFromLog, testCommitable
  )
where

import qualified List
import Maybe (fromJust, isNothing, isJust, maybeToList)

import Database.TRANSDB.Backend
import Database.TRANSDB.Manager
import qualified Database.TRANSDB.SQL as SQL

import qualified Database.HDBC as DB


-- | Return values from DB. The @Maybe@ type distinguishes the NULL value from
-- other values. 
type DBValue = SQL.DBValue
-- | One row of a tuple value returned from DB.
type NamedValue = SQL.NamedValue
-- | Conjuncts of a where clause
type WhereClause = SQL.WhereClause

-- |prepare an sql statement and run it.
doSQL :: D -> String -> [DBValue] -> IO DB.Statement
doSQL db sql args =
  do conn <- getCachedConnection db
     st <- DB.prepare conn sql
     DB.sExecute st args
     return st

-- |prepare an sql statement and pass it to a wrapped function. captures errors
-- by casting 'SQLError' into a standard exception.
withSQL :: D -> String -> [DBValue] -> (DB.Statement -> IO a) -> IO a
withSQL db sql args fun =
  DB.handleSqlError $
  do conn <- getCachedConnection db
     st <- DB.prepare conn sql
     DB.sExecute st args
     r <- fun st
     DB.commit conn
     return r



-- |Internal use.
findIO :: D -> TableName -> WhereClause -> IO [PrimaryKey]
findIO db table whereClause =
  let pkName = primaryKeyColumn table
      sql = SQL.mkSelect [pkName] [table] whereClause 
  in
  withSQL db sql [] $ \ st ->
  do r <- DB.sFetchAllRows st
     let getPK [Just pk] = pk
     return $ map getPK r

-- |Check the fields of the named values for existence. Throws exception if the
-- fields are not available.
checkFieldsForExistence :: D -> TableName -> [NamedValue] -> IO ()
checkFieldsForExistence db table nvs =
  let sql = SQL.mkSelect (map fst nvs) [table] [] in
  withSQL db sql [] $ \ st ->
     return ()

-- |Handle for a transactional tuple. Contains the database handle, the name of
-- the table and the primary key value of the tuple.
data T =
  T { t_database :: D
    , t_tableName :: TableName
    , t_primaryKey :: PrimaryKey
    }
  deriving (Eq, Ord)

instance Show T where
  showsPrec _ (T db tn pk) =
    shows db .
    shows tn .
    shows pk

instance Read T where
  readsPrec i str =
    do (db, rest) <- reads str
       (tn, rest) <- reads rest
       (pk, rest) <- reads rest
       return (T db tn pk, rest)
    

-- |Obtain the result of a get request for a list of columns.
readValues :: T -> [ColumnName] -> [LogEntry] -> IO ([DBValue], [LogEntry])
readValues tid colNames log =
  do (nvs, entries) <- readValues' tid colNames log
     return (map snd nvs, entries)
readValues' :: T -> [ColumnName] -> [LogEntry] -> IO ([NamedValue], [LogEntry])
readValues' tid colNames log =
  g log
  where
    g [] =
      do namedValues <- reallyReadValues tid
	 let namedValues' = select colNames namedValues
	 return (namedValues', 
		 [GET_TUPLE tid namedValues'
		 ,READ_TUPLE tid namedValues [] []])
    g (READ_TUPLE { le_tid = tid', le_nvs = namedValues } : rest) | tid == tid' =
      let namedValues' = select colNames namedValues in
      return (namedValues', [GET_TUPLE tid namedValues'])
    g (NEW_TUPLE tid' namedValues : rest) | tid == tid' =
      let namedValues' = select colNames namedValues in
      return (namedValues', [GET_TUPLE tid namedValues'])
    g (REM_TUPLE tid' : rest) | tid == tid' =
      error ("DB.Transaction.readValues: attempt to access removed tuple")
    g (SET_TUPLE tid' namedValues : rest) | tid == tid' =
      let colsInSet = map fst namedValues
	  colsFromSet = List.intersect colNames colsInSet
	  colsFromRest = (List.\\) colNames colsInSet
	  selectedFromSet = select colsFromSet namedValues 
      in
      if null colsFromRest
      then return (selectedFromSet, [GET_TUPLE tid selectedFromSet])
      else if null colsFromSet
      then g rest
      else do
	-- recursive call requires/guarantees that all asked columns are present
        -- in the result
        (selectedFromRest, GET_TUPLE _ _ : logEntries) <-
          readValues' tid colsFromRest rest
	let namedValues' = merge colNames selectedFromSet selectedFromRest
	return (namedValues', GET_TUPLE tid namedValues' : logEntries)
    g (_ : rest) =
      g rest

-- |Obtain the result of a @get@ request for all values.
readAllValues :: T -> [LogEntry] -> IO ([NamedValue], [LogEntry])
readAllValues tid log =
  g log
  where
    g [] =
      do namedValues <- reallyReadValues tid
	 return (namedValues, 
		 [GET_TUPLE tid namedValues,
		  READ_TUPLE tid namedValues [] []])
    g (READ_TUPLE { le_tid = tid', le_nvs = namedValues } : rest) | tid == tid' =
      return (namedValues, [GET_TUPLE tid namedValues])
    g (NEW_TUPLE tid' namedValues : rest) | tid == tid' =
      return (namedValues, [GET_TUPLE tid namedValues])
    g (SET_TUPLE tid' namedValues' : rest) | tid == tid' =
      do (namedValues, GET_TUPLE _ _ : logEntries) <- g rest
	 let namedValues'' = SQL.update namedValues namedValues'
	 return (namedValues'', GET_TUPLE tid namedValues'' : logEntries)
    g (_ : rest) =
      g rest

-- |Reads a full tuple identified by its primary key.
reallyReadValues :: T -> IO [NamedValue]
reallyReadValues (T db table pk) =
  let sql = SQL.mkSelect ["*"] [table]
              $ SQL.mkWhereEqual [primaryKeyColumn table]
  in
  withSQL db sql [Just pk] $ \ st ->
  do r <- DB.sFetchRow st
     colNames <- DB.getColumnNames st
     case r of
       Just values ->
	 return (zip colNames values)
       Nothing ->
	 return (map (\colName -> (colName, Nothing)) colNames)

reallyReadTuples ::
  D -> TableName -> WhereClause -> [NamedValue] -> String -> IO [LogEntry]
reallyReadTuples db table whereClause nvs extraSQL =
  let sql = (SQL.mkSelect ["*"] [table]
              $ (SQL.mkWhereEqualNULL nvs ++ whereClause))
	    ++ extraSQL
      mkREAD nvs' =
	let Just (Just pk) = lookup (primaryKeyColumn table) nvs' in
	READ_TUPLE { le_tid = T db table pk
		   , le_nvs = nvs'
		   , le_where = whereClause
		   , le_selnvs = nvs }
  in
  withSQL db sql [] $ \st ->
  do r <- DB.sFetchAllRows st
     d <- DB.describeResult st
     return (map (mkREAD . zip (map fst d)) r)

-- |Checks existence of tuple by querying the db
reallyCheckExists :: T -> IO Bool
reallyCheckExists (T db table pk) =
  let pkColumn = primaryKeyColumn table
      sql = SQL.mkSelect ["COUNT(*)"] [table]
              $ SQL.mkWhereEqual [pkColumn]
  in
  withSQL db sql [Just pk] $ \st ->
  do r <- DB.sFetchRow st
     case r of
       Just [Just thecount] ->
	 return (read thecount > 0)
       _ ->
	 return False

-- 
type AccessedNames = [ColumnName]
type DatabaseName = String
data LogEntry =
    WITH_TRANSACTION
  | NEW_TUPLE { le_tid :: T
	      , le_nvs :: [NamedValue]    -- ^ results from intial write
	      }
    -- ^ asserts that the tuple did not exist before
  | READ_TUPLE { le_tid :: T
	       , le_nvs :: [NamedValue]	    -- ^ results from initial read; 
	       , le_where ::		    -- select condition must still be true
		     WhereClause	    -- free form condition
	       , le_selnvs ::
		     [NamedValue]	    -- selection by name=value
	       } -- ^asserts that the tuple did exist
  | END_TUPLE 	 -- ^ marks the end of a list of @READ_TUPLE@s
  | GET_TUPLE { le_tid :: T
	      , le_nvs :: [NamedValue]
	      }		    -- ^ record of a get op; the values are not really important 
  | SET_TUPLE { le_tid :: T
	      , le_nvs :: [NamedValue]
	      }             -- ^ record of a set op
  | REM_TUPLE { le_tid :: T
	      }		    -- ^ asserts that the tuple did exist


eitherToMaybe :: Either a b -> Maybe b
eitherToMaybe (Left x) = Nothing
eitherToMaybe (Right x) = Just x

instance Show LogEntry where
  showsPrec _ WITH_TRANSACTION =
    showChar 'W'
  showsPrec _ END_TUPLE =
    showChar 'E'
  showsPrec _ (REM_TUPLE tid) =
    showChar 'D' . shows tid
  showsPrec _ entry =
    (case entry of
      NEW_TUPLE {} ->
	showChar 'N'
      READ_TUPLE {} ->
	showChar 'R'
      GET_TUPLE {} ->
	showChar 'G'
      SET_TUPLE {} ->
	showChar 'S')
    . shows (le_tid entry) 
    . showsNamedValues (le_nvs entry)
    . case entry of 
	READ_TUPLE {} ->
	  shows (le_where entry) . shows (le_selnvs entry)
	_ ->
	  id

showsNamedValues [] =
  showChar '.'
showsNamedValues (nv : nvs) =
  showChar ',' .
  showsNamedValue nv .
  showsNamedValues nvs

showsNamedValue (name, value) =
  shows name .
  showsValue value

showsValue Nothing =
  showChar '?'
showsValue (Just s) =
  showChar '!' .
  shows s

instance Read LogEntry where
  readsPrec i str =
    case str of
      'W' : rest ->
	return (WITH_TRANSACTION, rest)
      'E' : rest ->
	return (END_TUPLE, rest)
      'D' : rest ->
	do (tid, rest) <- reads rest
	   return (REM_TUPLE tid, rest)
      'N' : rest ->
	cont NEW_TUPLE rest
      'R' : rest ->
	do (mkEntry, rest) <- cont READ_TUPLE rest
	   (whereClause, rest) <- reads rest
	   (selNvs, rest) <- reads rest
	   return (mkEntry whereClause selNvs, rest)
      'G' : rest ->
	cont GET_TUPLE rest
      'S' : rest ->
	cont SET_TUPLE rest
      _ ->
	fail "unable to read database log entry"
    where
      cont mkEntry str =
	do (tid, str) <- reads str
	   (nvs, str) <- readsNamedValues str
	   return (mkEntry tid nvs, str)

readsNamedValues str =
  case str of
    '.' : rest ->
      return ([], rest)
    ',' : rest ->
      do (nv, rest) <- readsNamedValue rest
	 (nvs, rest) <- readsNamedValues rest
	 return (nv : nvs, rest)

readsNamedValue str =
  do (name, str) <- reads str
     (value, str) <- readsValue str
     return ((name, value), str)

readsValue str =
  case str of
    '?' : rest ->
      return (Nothing, rest)
    '!' : rest ->
      do (str, rest) <- reads rest
	 return (Just str, rest)

-- |Access descriptor which summarizes accesses to one database tuple.
data TUP_DESC =
    TUP_DESC 
      { tup_tid :: T
      , tup_selected :: WhereClause
      , tup_selnvs   :: [NamedValue]
      , tup_read :: Maybe [NamedValue]
			-- ^ @Nothing@: created,
			--   @Just nvs@ result of initial read
      , tup_accessed :: [ColumnName]
			-- ^ accesses to initial values
      , tup_write :: Maybe [NamedValue]
			-- ^ @Nothing@: removed,
	                --   @Just nvs@ akkumulated writes
      }
    deriving (Show)

-- |Process the log to compile overall access descriptors for committing.
getDescriptors :: [LogEntry] -> [TUP_DESC]
getDescriptors =
  foldr f []
  where
    f :: LogEntry -> [TUP_DESC] -> [TUP_DESC]
    f (NEW_TUPLE tid nvs) tds =
      TUP_DESC  { tup_tid = tid
		, tup_selected = []
		, tup_selnvs = []
		, tup_read = Nothing
		, tup_accessed = []
		, tup_write = Just nvs } : tds
    f (READ_TUPLE tid nvs wcl selnvs) tds =
      TUP_DESC  { tup_tid = tid
		, tup_selected = wcl
		, tup_selnvs = selnvs
		, tup_read = Just nvs
		, tup_accessed = []
		, tup_write = Just [] } : tds
    f END_TUPLE tds = 
      tds
    f (REM_TUPLE tid) tds =
      transform tid tds
	(\ td -> td { tup_write = Nothing })
    f (GET_TUPLE tid nvs) tds =
      transform tid tds
	(\ td -> let accessed_earlier = tup_accessed td
		     accessed_now = map fst nvs
		     overwritten = map fst (fromJust $ tup_write td) in
	         td { tup_accessed = accessed_earlier
		        ++ (accessed_now List.\\ (overwritten ++ accessed_earlier)) })
    f (SET_TUPLE tid nvs) tds =
      transform tid tds
	(\ td -> td { tup_write = Just (overwrite (fromJust $ tup_write td) nvs) })
    transform tid tds f =
      case tds of
	[] ->
	  error "DB.Transaction.getDescriptors: trying to update nonexistent tuple descriptor"
	td : tds ->
	  if tup_tid td == tid
	  then f td : tds
	  else td : transform tid tds f

-- | Get the descriptors and try to commit
commitFromLog :: [LogEntry] -> IO Bool
commitFromLog les =
  do let des = getDescriptors les
     tryToCommit des

-- | Test if the accumulated @TUP_DESC@ records are commitable to the database
testCommitable :: [LogEntry] -> IO Bool
testCommitable les =
  let des = getDescriptors les in
  checkAccessedValuesUnchanged des

-- | Try to commit the accumulated @TUP_DESC@ records to the database.
tryToCommit :: [TUP_DESC] -> IO Bool
tryToCommit tds =
  let commit =
        do conns <- getAllConnections tds
	   mapM_ DB.commit conns
      rollback =
	do conns <- getAllConnections tds
	   mapM_ DB.rollback conns
  in
  -- start transaction with the database; implicit in @HDBC@
  -- what about error handling?
  if readOnlyTransaction tds
  then do commit
	  return True
  else DB.handleSql (\sqlerr -> do { rollback; return False }) $ do
     commit
     accessedValuesUnchanged <- checkAccessedValuesUnchanged tds
     if accessedValuesUnchanged
       then do writeSuccessful <- writeChangedTuples tds
	       if writeSuccessful then commit else rollback
	       return writeSuccessful
       else do rollback		-- nothing to rollback, anyway (nothing written)
	       return False

-- | Retrieves all connections involved in a list of 'TUP_DESC' records from the
-- cache. Returns it as a list sorted by the 'D' component.
getAllConnections :: [TUP_DESC] -> IO [DB.Connection]
getAllConnections tds =
  let allDbs = removeDuplicates (List.sort (map (t_database . tup_tid) tds))
      removeDuplicates (x : rest1@(y : _)) =
	if x == y then removeDuplicates rest1
		  else x : removeDuplicates rest1
      removeDuplicates xs = 
	xs
  in
  mapM getCachedConnection allDbs

-- | Reread all tuples and check that their accessed values have not
-- changed. Only reads up to first change.
-- To be called inside a transaction, so it does not commit.
checkAccessedValuesUnchanged :: [TUP_DESC] -> IO Bool
checkAccessedValuesUnchanged [] =
  return True
checkAccessedValuesUnchanged (td : tds) =
  do unchanged <- tupleUnchanged td
     if unchanged 
	then checkAccessedValuesUnchanged tds
	else return False
  where
    tupleUnchanged tup =
	 -- Read with 
	 -- SELECT * FROM table WHERE pk=xxx, c1=previous1, ... , cn=previousn
	 -- return True if we get a tuple, False otherwise
      let tid = tup_tid tup
	  table = t_tableName tid
	  pkColumn = primaryKeyColumn table
	  pkValue  = Just (t_primaryKey tid)
	  nvs = (pkColumn, pkValue) : columnsRead tup
	  mkSelect s = SQL.mkSelect [pkColumn] [table] s
	  sql = mkSelect (SQL.mkWhereEqualNULL (nvs ++ tup_selnvs tup)
			  ++ tup_selected tup)
      in
      do st <- doSQL (t_database tid) sql []
	 r <- DB.sFetchRow st
	 case tup_read td of
	   Nothing ->
	     -- we created the tuple, so check that it still does not exist!
	     return (isNothing r)
	   Just _ ->
	     -- tuple is unchanged if result is Just something
	     return (isJust r)

-- |Determine the columns which have been read together with their initial
-- values.
columnsRead :: TUP_DESC -> [NamedValue]
columnsRead td =
  case tup_read td of
    Nothing ->
      []
    Just nvs ->
      select (tup_accessed td) nvs

-- |Identifies those tuples that have changed and writes them back to the
-- database. Could be made more efficient by first grouping by database and
-- table and then executing all updates at once.
writeChangedTuples :: [TUP_DESC] -> IO Bool
writeChangedTuples tds =
  let actions = prepareAllWriteOperations tds
      g [] = 
	return True
      g (iob : iobs) =
	do b <- iob
	   if b then g iobs else return False
  in g actions

-- |Determines the required write operations and schedules them in order INSERT,
-- UPDATE, DELETE.
prepareAllWriteOperations :: [TUP_DESC] -> [IO Bool]
prepareAllWriteOperations tds =
  let ops = map prepareWriteOperation tds
      opsIUD = filter (/= OpNothing) ops
      opsSorted = List.sort opsIUD
      doOp op =
	do conn <- getCachedConnection (op_db op)
	   st <- DB.prepare conn (op_sql op)
	   nr <- DB.sExecute st (op_parms op)
	   return (nr == 1)
  in
    map doOp opsSorted

-- | description of an operation
data Operation =
    OpInsert  { op_db :: D
	      , op_sql :: String
	      , op_parms :: [DBValue]
	      }
  | OpUpdate  { op_db :: D
	      , op_sql :: String
	      , op_parms :: [DBValue]
	      }
  | OpDelete  { op_db :: D
	      , op_sql :: String
	      , op_parms :: [DBValue]
	      }
  | OpNothing
  deriving (Ord, Eq)

-- |Attempt to write a tuple. Returns @True@ if successful.
prepareWriteOperation :: TUP_DESC -> Operation
prepareWriteOperation td =
  let tid = tup_tid td
      db = t_database tid
      tableName = t_tableName tid
      pkColumn = primaryKeyColumn tableName
      pkValue  = Just (t_primaryKey tid)
      mnvs = tup_write td
  in
  case mnvs of
    Nothing | isJust (tup_read td) ->		-- remove
      let sql = SQL.mkDelete (t_tableName tid)
	      $ SQL.mkWhereEqual [pkColumn] in
      OpDelete  { op_db = db
		, op_sql = sql
		, op_parms = [pkValue]
		}
    Just (nvs@(_:_)) | isJust (tup_read td) ->  -- update old entry
      let sql = SQL.mkUpdate (t_tableName tid) nvs 
	      $ SQL.mkWhereEqual [pkColumn] in
      OpUpdate  { op_db = db
		, op_sql = sql
		, op_parms = [pkValue]
		}
    Just (nvs@(_:_)) | isNothing (tup_read td) -> -- new tuple
      let columns = pkColumn : map fst nvs
	  values  = pkValue  : map snd nvs
	  sql = SQL.mkInsert (t_tableName tid) columns in
      OpInsert  { op_db = db
		, op_sql = sql
		, op_parms = values
		}
    _ ->  	-- do nothing for @Just []@ and @Nothing@\/@Nothing@
      OpNothing


-- | Check that a transaction description does not contain any write
-- operations. 
readOnlyTransaction :: [TUP_DESC] -> Bool
readOnlyTransaction tds =
  and (map isReadOnlyTUP tds)
  where
    isReadOnlyTUP tup = 
      tup_write tup == Just []

----------------------------------------------------------------------
-- list utilities

-- |Select named columns from a row of 'NamedValue'.
select :: (Eq a) => [a] -> [(a, b)] -> [(a, b)]
select colNames namedValues =
  map getNamedValue colNames
  where
    getNamedValue colName =
      (colName, fromJust (lookup colName namedValues))

-- |Merge two rows and project given column names.
merge :: (Eq colName) => [colName] -> [(colName, b)] -> [(colName, b)] -> [(colName, b)]
merge colNames namedColsHere namedColsThere =
  let namedColsEverywhere = namedColsHere ++ namedColsThere in
  select colNames namedColsEverywhere


-- |Overwrite some columns by new columns.
overwrite :: (Eq colName) => [(colName, a)] -> [(colName, a)] -> [(colName, a)]
overwrite tup tup_upd =
  foldr one_upd tup tup_upd
  where
    one_upd tup_upd@(col_upd, val_upd) tup =
      g tup
      where
        g [] =
	  [tup_upd]
	g ((col, val) : rest) =
	  if col == col_upd
	  then (col, val_upd) : rest
	  else (col, val) : g rest
