--  2006 Peter Thiemann
{- | Database connection manager.
-}
module Database.TRANSDB.Manager
  (
    D (), register, registerIO,
    getCachedConnection
  )
where

import System.IO.Unsafe (unsafePerformIO)
import Data.IORef
import Maybe (isNothing)
import Monad (when, mplus)

import qualified Database.HDBC as DB

-- | Abstract type of database handles. Contains a hashvalue for the db.
-- 
-- A database handle does not contain the connection itself because there is no
-- safe way to reconstruct it from the log.
newtype D =
  D String
  deriving (Eq, Ord)

instance Show D where
  showsPrec _ (D hash) =
    shows hash

instance Read D where
  readsPrec i str =
    do (hash, rest) <- reads str
       return (D hash, rest)

-- | Create a hash value.
makeD :: String -> D
makeD = D

-- | Register a database for use in a transaction. Takes a connection specifier and
-- a function that creates the connection.
-- Returns a database handle. Uses the showed rendition of the connection
-- specifier as identification. Creates the connections on demand and caches them.
-- 
-- Example usage: @register \"dbname=exampledb\" DB.connectPostgreSQL@
-- 
-- Must be @MVar@ed or @TVar@ed in a server setting.
register :: (Show cs) => cs -> (cs -> IO DB.Connection) -> D
register connectionSpec connect =
  unsafePerformIO $ registerIO connectionSpec connect

-- | Internal use only.
-- 
-- Must be synchronized in a concurrent setting.
registerIO :: (Show cs) => cs -> (cs -> IO DB.Connection) -> IO D
registerIO connectionSpec connect =
  do let hash = makeD (show connectionSpec)
     alist <- readIORef connectionCache
     when (isNothing (lookup hash alist)) $
       writeIORef 
	 connectionCache
	 ((hash, unsafePerformIO (connect connectionSpec)) : alist)
     return hash


-- | Retrieve a cached connection. Internal use.
getCachedConnection :: D -> IO DB.Connection
getCachedConnection db =
  if db /= D "" then {- always @True@, just to force evaluation of @db@ -}
  do alist <- readIORef connectionCache
     case lookup db alist of
       Just conn ->
	    return conn
       Nothing ->
	    error ("getCachedConnection failed to find " ++ show db ++
		". Cache contains keys: " ++ show (map fst alist))
  else error ("getCachedConnection is severly mixed up")

-- | Association list from database handles to db connections. 
{-# NOINLINE connectionCache #-}
connectionCache :: IORef [(D, DB.Connection)]
connectionCache = unsafePerformIO (newIORef [])
