--  2006 Peter Thiemann
{- | Convenience functions for constructing SQL query strings and manipulating
-- query results.
-}
module Database.TRANSDB.SQL
  ( 
  -- * Types
  ColumnName, TableName, DBValue, NamedValue, WhereClause,
  -- * SQL construction
  mkInsert, mkSelect, mkUpdate, mkDelete, mkWhereEqual, mkWhereEqualNULL,
  -- * Helpers
  select, merge, update
  )
where

import Char
import List
import Maybe

type ColumnName = String
type TableName = String

-- |A where clause is a list of conjuncts
type WhereClause = [String] 
    
-- | Return values from DB. The @Maybe@ type distinguishes the NULL value from
-- other values. 
type DBValue = Maybe String
-- | One row of a tuple value returned from DB.
type NamedValue = (ColumnName, DBValue)

-- |Construct an @INSERT@ query template from column names. Values must be
-- supplied on execute.
mkInsert :: TableName -> [ColumnName] -> String
mkInsert tableName colNames =
  "INSERT INTO " ++ showName tableName ++
  nameList colNames ++
  "VALUES" ++ placeholderList (map (const "?") colNames)

-- |Construct a @SELECT@ query from column expressions, table names, and a where
-- clause string.
-- 
-- Attn: expects to select column names
mkSelect :: [ColumnName] -> [TableName] -> WhereClause -> String
mkSelect colExprs tableNames whereClause =
  "SELECT " ++ list2string colExprs ++
  " FROM " ++ list2string tableNames ++
  mkCondition whereClause

-- |Construct an @UPDATE@ query from table name, the new assignment, and a where
-- clause string.
mkUpdate :: TableName -> [NamedValue] -> WhereClause -> String
mkUpdate tableName nvs whereClause =
  "UPDATE " ++ showName tableName ++
  " SET " ++ showNamedValues nvs ++
  mkCondition whereClause

-- |Construct a @DELETE@ query from a table name, and a where clause string.
mkDelete :: TableName -> WhereClause -> String
mkDelete tableName whereClause =
  "DELETE FROM " ++ showName tableName ++
  mkCondition whereClause

-- |construct a where clause from a list of conjuncts
mkCondition whereClause =
  g " WHERE " whereClause
  where
    g prefix [] = []
    g prefix (conj : conjs) = prefix ++ conj ++ g " AND " conjs

-- |Construct a @WHERE@ condition  of the form @col1=? AND col2=? ...@
mkWhereEqual :: [ColumnName] -> WhereClause
mkWhereEqual =
  map (\colName -> (showName colName ++ "=?"))
  
-- |Construct a @WHERE@ condition  of the form @col1=? AND col2=? ...@ which
-- works correctly with @NULL@
mkWhereEqualNULL :: [NamedValue] -> WhereClause
mkWhereEqualNULL =
  map g
  where
    g (colName, value) =
      showName colName ++ equalTo value
    equalTo Nothing = 
      " IS NULL"
    equalTo (Just s) =
      '=' : showQuoted s

-- |Display string in single quotes.
showQuoted :: String -> String
showQuoted s = '\'' : escape s ++ "'"

escape :: String -> String
escape = concatMap escapeSingle

escapeSingle :: Char -> String
escapeSingle x | isAlphaNum x = [x]
	       | otherwise    = '\\' : showOctal3 (ord x)

showOctal3 x = 
  let (d, r) = x `divMod` 64 in
  chr (ord '0' + d) : showOctal2 r
showOctal2 x =
  let (d, r) = x `divMod` 8 in
  chr (ord '0' + d) : showOctal1 r
showOctal1 x =
  [chr (ord '0' + x)]

-- |Transforms any string into a valid SQL identifier (aka name).
showName :: String -> String
showName s = '\"' : escapeName s ++ "\""

escapeName :: String -> String
escapeName = concatMap escapeNameSingle

escapeNameSingle :: Char -> String
escapeNameSingle '\"' = "\"\""
escapeNameSingle x = [x]

-- |Select named columns from a row of 'NamedValue'.
-- 
-- Specialized type @[ColumnName] -> [NamedValue] -> [NamedValue]@
select :: (Eq a) => [a] -> [(a, b)] -> [(a, b)]
select colNames namedValues =
  map getNamedValue colNames
  where
    getNamedValue colName =
      (colName, fromJust (lookup colName namedValues))

-- |Merge two rows and project given column names.
merge :: (Eq colName) => [colName] -> [(colName, b)] -> [(colName, b)] -> [(colName, b)]
merge colNames namedColsHere namedColsThere =
  let namedColsEverywhere = namedColsHere ++ namedColsThere in
  select colNames namedColsEverywhere

-- |Merge two rows by overriding settings from the first row with settings from
-- the second row argument. Assume that second row updates a subset of the first
-- row.
update :: (Eq colName) =>
  [(colName, value)] -> [(colName, value)] -> [(colName, value)]
update oldBindings newBindings =
  map updateBinding oldBindings
  where
    updateBinding nv@(name, value) =
      case lookup name newBindings of
	Nothing ->
	  nv
	Just value' ->
	  (name, value')

showNamedValues :: [NamedValue] -> String
showNamedValues [] = 
  ""
showNamedValues (nv : nvs) =
  showNamedValue nv ++ if null nvs then "" else ',' : showNamedValues nvs

showNamedValue :: NamedValue -> String
showNamedValue (name, value) =
  showName name ++ '=' : showValue value

showValue :: DBValue -> String
showValue Nothing = "NULL"
showValue (Just v) = showQuoted v

list2string [] = ""
list2string (x : xs) =
  showName x ++ if null xs then "" else ',' : list2string xs

parenList showsElem xs = 
  g '(' xs ")"
  where 
  g pre xs =
    showChar pre .
    case xs of
      [] ->
        id
      x : xs ->
	showsElem x . h ',' xs
  h pre xs =
    case xs of
      [] ->
        id
      x : xs ->
	showChar pre . showsElem x . h pre xs

nameList =
  parenList (\n -> (showName n ++))

valueList =
  parenList showString

placeholderList =
  parenList showString
