--  2006 Peter Thiemann
{- | String-based, direct-style implementation of transactional database access.
-}
module Database.TRANSDB.Transaction 
  (
    -- * Database connection
    D (), register,
    -- * Database tuple handles
    T (), findByPK, findByNV, find, getPrimaryKey,
    -- * Operations on tuples
    create, remove, 
    get, getMany, getAll, set, setMany, 
    -- * Transaction support
    with, orElse, abandon, TDB ()
  )
where

-- import qualified Database.TRANSDB.TraceQuery as DB

import qualified List
import Maybe (fromJust)
import IO


import Database.TRANSDB.Backend
import Database.TRANSDB.Manager
import Database.TRANSDB.LowLevel

import qualified Database.TRANSDB.SQL as SQL


-- |The transactional database monad is a state monad lifted over @IO@ where the
-- state is the transaction log. Not using the monad transformer @StateT@ to
-- avoid dependencies on type system extensions like
-- functional parameters etc. Compared to @Direct@, there is an additional
-- reader component which captures the current bind-continuation. The additional
-- argument is the toplevel result. Usually @r=b@.
-- There also the error monad @Either@ mixed in to indicate whether a
-- transaction is abandoned.
data TDB r a =
    TDB { unTDB :: TState r -> (a -> TDB r r) -> IO (Either r a, TState r)}

data TState r =
    TState { getLog :: [LogEntry]
	   , getCPs :: [(TState r, TDB r r)] 	-- ^stack of choice points
	   }

instance Monad (TDB r) where
  return x = TDB (\ s c -> return (Right x, s))
  m >>= f  = TDB (\ s c -> do (x', s') <- unTDB m s (composeM c f)
			      case x' of
				Left xl' ->
				  return (Left xl', s')
				Right xr' ->
				  unTDB (f xr') s' c)

ioTDB :: IO a -> TDB r a
ioTDB action
  = TDB (\ s c -> do x <- action
		     return (Right x, s))
  
-- |Kleisli composition
composeM :: (Monad m) => (y -> m z) -> (x -> m y) -> (x -> m z)
composeM c f =
  \ x -> f x >>= c

-- |Install a choice point by applying the second argument to its continuation,
-- storing that, and starting to execute the first argument.
-- @
-- orElse :: TDB r' x -> TDB r x -> TDB r x
-- @
-- The above type is correct but cannot be checked. Intuitively, the @orElse@
-- makes sure that the exceptions generated by its first argument never get
-- through. 
orElse :: TDB r x -> TDB r x -> TDB r x
orElse t1 t2 =
  TDB (\ s c -> 
   do let s' = s { getCPs = (s, t2 >>= c) : getCPs s } 
      (x'', s'') <- unTDB t1 s' c
      case x'' of
	Left xl'' ->
	  unTDB t2 s c
	Right xr'' ->
	  return (x'', s'')
      )

-- |Abandons the innermost active transaction and makes it return its @result@
-- argument. May be captured with an `orElse`.
abandon :: result -> TDB result x
abandon res =
  TDB (\ s c -> return (Left res, s))

-- |Start the action with the identity arrow @return@ of the Kleisli category as
-- its continuation. 
runTDB :: TState r -> TDB r r -> IO (Either r r, TState r)
runTDB ts tdb = 
  unTDB tdb ts return

-- |Apply a function to the log.
applyToLogTDB :: ([LogEntry] -> [LogEntry]) -> TDB r [LogEntry]
applyToLogTDB g =
  TDB (\ ts c -> 
       let x = getLog ts 
	   y = g x
       in  return (Right y, ts { getLog = y }))

getTheLog :: TMonad t => t [LogEntry]
getTheLog = 
  applyToLog id

-- |Apply a function to the stack of choice points
applyToCPsTDB :: ([(TState r, TDB r r)] -> [(TState r, TDB r r)]) ->
		 TDB r [(TState r, TDB r r)]
applyToCPsTDB g =
  TDB (\ ts c -> 
       let x = getCPs ts 
	   y = g x
       in  return (Right y, ts { getCPs = y }))

getTheCPs :: TDB r [(TState r, TDB r r)]
getTheCPs =
  applyToCPsTDB id

-- |A transaction monad @TMonad@ can access @IO@ and it has a log.
class Monad t => TMonad t where
  io         :: IO a -> t a
  applyToLog :: ([LogEntry] -> [LogEntry]) -> t [LogEntry]

instance TMonad (TDB r) where
  io = ioTDB
  applyToLog = applyToLogTDB


-- |Create a fresh tuple in @table@ with new primary key and set its initial
-- contents. Returns a handle for the tuple. The transaction will not commit if
-- the initial contents do not fit (but it should better raise an exception
-- immediately). 
-- 
-- Complication: the fresh tuple must not be visible outside this transaction
-- but we need to ensure that its unique primary key is reserved, anyway. Hence,
-- we just reserve it through the sequence first time round.
create :: D -> TableName -> [NamedValue] -> TDB r T
create db table namedValues =
  do tid <- io $ do 
        conn <- getCachedConnection db
	pk <- generatePrimaryKey conn table
	return $ T db table pk
     applyToLog (NEW_TUPLE tid (List.sort namedValues) :)
     return tid


-- |Remove tuple identified by a primary key from DB. Only registers the action
-- in the log, but checks that the tuple exists.
-- Fails (with an exception) if the tuple does not exist.
remove :: T -> TDB r ()
remove tid =
  do tupleExists <- checkTupleExists tid False
     if tupleExists 
	then 
	  do applyToLog (REM_TUPLE tid :)
	     return ()
	else
	  fail ("DB.Transaction.remove: tuple " ++ show tid ++ " not present")

-- |Check existence of a tuple.
checkTupleExists :: T -> Bool -> TDB r Bool
checkTupleExists tid doRead =
  do log <- getTheLog
     let tupleIDStatus = checkTupleIDStatus tid log
     case tupleIDStatus of
       Exists ->
	 return True
       Removed ->
	 return False
       Unknown ->
	 if doRead 
	 then
	   do nvs <- io $ reallyReadValues tid
	      let pkColumn = primaryKeyColumn (t_tableName tid)
	      case lookup pkColumn nvs of
		Just (Just _) ->
		  do applyToLog (READ_TUPLE tid nvs [] [] :)
		     return True
		_ ->
		  return False
	 else
	   do exists <- io $ reallyCheckExists tid
	      return exists

data TupleIDExistence =
    Exists | Removed | Unknown

-- |does this tuple ID exist in the log?
checkTupleIDStatus :: T -> [LogEntry] -> TupleIDExistence
checkTupleIDStatus tid les =
  foldr g Unknown les
  where
    g :: LogEntry -> TupleIDExistence -> TupleIDExistence
    g le b =
      if le_tid le /= tid 
      then b
      else -- this log entry is for our tuple
      case le of
	REM_TUPLE {} ->
	  Removed
	_ ->
	  Exists

-- |Obtain a tuple through its primary key. Checks for its existence and
-- preloads the tuple in the cache.
findByPK :: D -> TableName -> PrimaryKey -> TDB r (Maybe T)
findByPK db table pk =
  let tid = T db table pk in
  checkTupleExists tid True >>= \ tupleExists ->
  if tupleExists 
  then return (Just tid)
  else return Nothing
  
-- |Obtain tuples through a where-clause of the form
-- name_i=value_i. Preloads. Caution: may generate an unlimited number of log
-- entries. 
-- 
-- TODO: Should perhaps take a @LIMIT@ parameter
findByNV :: D -> TableName -> [NamedValue] -> String -> TDB r [T]
findByNV db table nvs extraSQL =
  do allReads <- io $ reallyReadTuples db table [] nvs extraSQL
     applyToLog (allReads ++)
     return (map le_tid allReads)

-- |Take a database, a table, and an SQL where-clause. 
-- Return a list of tuple IDs for the results.
find :: D -> TableName -> WhereClause -> TDB r [T]
find db table whereClause =
  do pks <- io (findIO db table whereClause)
     let makeT pk = T db table pk
     return $ map makeT pks

-- |Get the primary key from a tuple handle. Does not access the DB at all. Does
-- not affect the transactional behavior of the tuple (it does not count as a
-- read operation on the tuple).
getPrimaryKey :: T -> DBValue
getPrimaryKey tid =
  Just (t_primaryKey tid)

-- |Get field value from handle. Caches the whole tuple.
get :: T -> ColumnName -> TDB r (DBValue)
get tid field =
  do [value] <- getMany tid [field]
     return value

-- |Get field values from handle. Caches the whole tuple.
getMany :: T -> [ColumnName] -> TDB r [DBValue]
getMany tid colNames =
  do nvs <- getInternal tid (readValues' tid colNames)
     return (map snd nvs)

getInternal :: 
  T -> ([LogEntry] -> IO ([NamedValue], [LogEntry])) -> TDB r [NamedValue]
getInternal tid readValues =
  do log <- getTheLog
     (values, newEntries) <- io $ readValues log
     applyToLog (newEntries ++)
     return values

-- |Get whole tuple from handle.
getAll :: T -> TDB r [NamedValue]
getAll tid =
  getInternal tid (readAllValues tid)

-- |Set field value through handle.
set :: T -> ColumnName -> DBValue -> TDB r ()
set handle field val =
  -- trivial implementation
  setMany handle [(field, val)]

-- |Set many columns at once through handle
setMany :: T -> [NamedValue] -> TDB r ()
setMany tid namedValues =
  do applyToLog (SET_TUPLE tid namedValues :)
     return ()

-- |Abstract of the nested vs the unnested case. @target@ is the target
-- monad. It may be @IO@ (the unnested case) or @TDB r@ (the nested case).
class Monad target => With target where
  with :: 
     result				-- default result
     -> TDB result result
     -> target result

instance With IO where
  with = withIO

instance With (TDB r) where
  with = withTDB

-- |Opens a top-level transactional scope. Takes a default @result@
-- value and a transaction with the same @result@ type and constructs an @IO@
-- action returning the same value of the transaction (if that commits
-- successfully). The default value is returned if the transaction fails without
-- explicitly returning a result value (via @abandon@).
withIO :: result -> TDB result result -> IO result
withIO result body =
  let initialstate = 
	  TState { getLog = []
		 , getCPs = []
		 }
      loop state body =
	do (lr, finalstate) <- runTDB state body
	   let log = getLog finalstate
	       cps = getCPs finalstate
	       next result =
		 case cps of
		   (laststate, t) : _ ->
		      loop laststate t
		   [] ->
		      -- at this point, STM would wait for the world to change 
		      -- and retry. Instead we stop with
		      return result

	   case lr of
	     Right v ->
	       -- got all through, now try to commit
	       do committed <- commitFromLog log
		  if committed then
		     return v
		   else 
		     next result

	     Left err ->
		 -- finally @abandon@ed operation: @finalstate@ must be empty!
	       next err
  in
     loop initialstate body

-- |Opens a nested transactional scope. Same functionality as 'withIO'. However,
-- a nested transaction can only locally check that its preconditions hold. This
-- local check does not preclude that local consistency prevails until the
-- surrounding transaction commits. However, it is not possible to backtrack
-- into the nested transaction from the surrounding transaction, whereas the
-- surrounding transaction may attempt further choice points. A successful
-- nested transaction only manipulates the log.
withTDB :: a -> TDB a a -> TDB b a
withTDB result body =
  do outerLog <- getTheLog
     outerCPs <- getTheCPs
     let initialstate =
	   TState { getLog = outerLog
		  , getCPs = []
		  }
	 loop state body = do
	   (lr, finalstate) <- io $ runTDB state body
	   let log = getLog finalstate
	       cps = getCPs finalstate
	       next result =
		 case cps of
		   (laststate, t) : _ ->
		      loop laststate t
		   [] ->
		      -- at this point, STM would wait for the world to change 
		      -- and retry. Instead we stop with
		      return result

	   case lr of
	     Right v ->
	       -- succeed if current log state is consistent with the DB
	       do commitable <- io $ testCommitable log
		  if commitable then
		     do applyToLog (log ++)
			return v
		   else
		     -- otherwise, continue at the last choicepoint
		     next result
	     Left err ->
	       -- `abandon` can be captured by `orElse`
	       next err

     loop initialstate body
